from __future__ import annotations

import logging
import socket
import socketserver
import threading
import uuid
from threading import Thread
from time import time
from typing import Dict, Set, Tuple, cast

from ..comm.comm_buffer import CommBuffer
from ..protocol.command_req import CommandReq
from ..protocol.constants import DEFAULT_SERVER_PORT, PROGRAM_NAME, CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1SyncCommandEnum

log = logging.getLogger(__name__)

DISCONNECT_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.DISCONNECT).as_bytes
KEEP_ALIVE_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.KEEP_ALIVE).as_bytes
QUIT_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.QUIT).as_bytes
REBOOT_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.REBOOT).as_bytes
REGISTER_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.REGISTER).as_bytes
RESTART_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.RESTART).as_bytes
RESYNC_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.RESYNC).as_bytes
SENDING_STATE_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.SENDING_STATE).as_bytes
SHUTDOWN_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.SHUTDOWN).as_bytes
SYNC_BEGIN_RESPONSE: bytes = CommandReq(TMCC1SyncCommandEnum.SYNC_BEGIN).as_bytes
SYNC_COMPLETE_RESPONSE: bytes = CommandReq(TMCC1SyncCommandEnum.SYNC_COMPLETE).as_bytes
SYNC_STATE_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.SYNC_REQUEST).as_bytes
UPDATE_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.UPDATE).as_bytes
UPGRADE_REQUEST: bytes = CommandReq(TMCC1SyncCommandEnum.UPGRADE).as_bytes


class ProxyServer(socketserver.ThreadingTCPServer):
    __slots__ = "base3_addr", "ack", "dispatcher", "enqueue_proxy", "base3_dispatcher", "pdi_dispatcher"


class EnqueueProxyRequests(Thread):
    """
    Receives requests from PyTrain clients over TCP/IP and queues them for
    dispatch to the LCS SER2.
    """

    _instance = None
    _lock = threading.RLock()

    @classmethod
    def build(cls, buffer: CommBuffer, port: int = DEFAULT_SERVER_PORT) -> EnqueueProxyRequests:
        """
        Factory method to create an EnqueueProxyRequests instance
        """
        return EnqueueProxyRequests(buffer, port)

    @classmethod
    def clients(cls) -> Set[Tuple[str, int]]:
        if cls._instance is not None:
            return cls._instance.client_sessions
        raise AttributeError("EnqueueProxyRequests is not built yet.")

    @classmethod
    def stop(cls) -> None:
        if cls._instance is not None:
            cls._instance.shutdown()

    @classmethod
    def register_request(cls, port: int, client_id: uuid.UUID, version: tuple[int, int, int]) -> bytes:
        return cls._build_request(REGISTER_REQUEST, port, client_id, version)

    @classmethod
    def disconnect_request(cls, port, client_id: uuid.UUID) -> bytes:
        return cls._build_request(DISCONNECT_REQUEST, port, client_id)

    @classmethod
    def sync_state_request(cls, port: int = DEFAULT_SERVER_PORT, client_id: uuid.UUID = None) -> bytes:
        return cls._build_request(SYNC_STATE_REQUEST, port, client_id)

    @classmethod
    def sync_begin_response(cls) -> bytes:
        return SYNC_BEGIN_RESPONSE

    @classmethod
    def sync_complete_response(cls) -> bytes:
        return SYNC_COMPLETE_RESPONSE

    @classmethod
    def is_built(cls) -> bool:
        return cls._instance is not None

    @classmethod
    def server_port(cls) -> int:
        if cls._instance is not None:
            # noinspection PyProtectedMember
            return cls._instance._server_port
        raise AttributeError("EnqueueProxyRequests is not built yet.")

    @classmethod
    def server_ip(cls) -> int:
        if cls._instance is not None:
            # noinspection PyProtectedMember
            return cls._instance._server_ip
        raise AttributeError("EnqueueProxyRequests is not built yet.")

    @classmethod
    def _build_request(
        cls,
        request: bytes,
        port: int,
        client_id: uuid.UUID = None,
        version: tuple[int, int, int] = None,
    ) -> bytes:
        port = port if port else DEFAULT_SERVER_PORT
        client_id_bytes = client_id.bytes if client_id else bytes()
        if isinstance(version, tuple) and len(version) == 3:
            ver_bytes = bytes()
            for v in version:
                ver_bytes += v.to_bytes(1, "big")
        else:
            ver_bytes = bytes()
        return request + int(port & 0xFFFF).to_bytes(2, byteorder="big") + client_id_bytes + ver_bytes

    def __init__(self, tmcc_buffer: CommBuffer, server_port: int = DEFAULT_SERVER_PORT) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Enqueue Receiver")
        self._tmcc_buffer: CommBuffer = tmcc_buffer
        self._server_port = server_port
        self._server_ip = None
        self._lock = threading.RLock()
        self._clients: Dict[Tuple[str, int, uuid.UUID | None], float] = dict()
        self.start()

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in the system
        """
        with cls._lock:
            if EnqueueProxyRequests._instance is None:
                EnqueueProxyRequests._instance = super(EnqueueProxyRequests, cls).__new__(cls)
                EnqueueProxyRequests._instance._initialized = False
            return EnqueueProxyRequests._instance

    def client_connect(self, client_ip: str, port: int = DEFAULT_SERVER_PORT, client_id: uuid.UUID = None) -> None:
        # check if (client_ip, port) is unique, it should be unless
        # the previous client on this port unexpectedly disconnected
        with self._lock:
            disconnected = set()
            for k_ip, k_port, k_uuid in self._clients.keys():
                if k_ip == client_ip and k_port == port:
                    if client_id != k_uuid:
                        disconnected.add((k_ip, k_port, k_uuid))
            # delete disconnected key
            for k in disconnected:
                log.info(f"Purging disconnected client: {k}...")
                self._clients.pop(k, None)
            # record new client
            self._clients[(client_ip, port, client_id)] = time()

    def client_disconnect(self, client_ip: str, port: int = DEFAULT_SERVER_PORT, client_id: uuid.UUID = None) -> None:
        with self._lock:
            self._clients.pop((client_ip, port, client_id), None)

    def is_client(self, client_ip: str, port: int = DEFAULT_SERVER_PORT, client_id: uuid.UUID = None) -> bool:
        with self._lock:
            return (client_ip, port, client_id) in self._clients

    def client_alive(self, client_ip: str, port: int, client_id: uuid.UUID) -> None:
        with self._lock:
            if (client_ip, port, client_id) in self._clients:
                self._clients[(client_ip, port, client_id)] = time()
            else:
                log.error(f"Client {client_ip}:{port} is not registered, attempting restart...")
                if (client_ip, port) in self.client_sessions:
                    log.error(f"Can not restart client at {client_ip}:{port}; port in use")
                else:
                    from .command_listener import CommandDispatcher

                    CommandDispatcher.get().signal_clients(TMCC1SyncCommandEnum.RESTART, client_ip, port)

    def enqueue_request(self, data: bytes) -> None:
        self._tmcc_buffer.enqueue_command(data)

    @property
    def client_sessions(self) -> Set[Tuple[str, int]]:
        with self._lock:
            return {(k[0], k[1]) for k, v in self._clients.items()}

    def run(self) -> None:
        from .command_listener import CommandDispatcher
        from .. import get_version_bytes

        """
        Simplified TCP/IP Server listens for command requests from client and executes them
        on the PyTrain server.
        """
        # noinspection PyTypeChecker
        with ProxyServer(("", self._server_port), EnqueueHandler) as server:
            server.session_id = self._tmcc_buffer.session_id
            if self._tmcc_buffer.base3_address:
                server.base3_addr = self._tmcc_buffer.base3_address
                server.ack = get_version_bytes() + str.encode(server.base3_addr)
            else:
                server.ack = get_version_bytes()
            server.dispatcher = CommandDispatcher.get()
            server.base3_dispatcher = None
            server.pdi_dispatcher = None
            server.enqueue_proxy = self
            server.serve_forever()


class EnqueueHandler(socketserver.BaseRequestHandler):
    def __init__(self, request: socket.socket, client_address: Tuple[str, int], server: ProxyServer) -> None:
        super().__init__(request, client_address, server)

    def handle(self):
        from ..pdi.base3_buffer import Base3Buffer
        from ..pdi.constants import PDI_SOP, PdiCommand
        from ..pdi.pdi_listener import PdiDispatcher
        from .command_listener import CommandDispatcher

        byte_stream = bytes()
        ack = cast(ProxyServer, self.server).ack
        dispatcher: CommandDispatcher = cast(ProxyServer, self.server).dispatcher
        enqueue_proxy: EnqueueProxyRequests = cast(ProxyServer, self.server).enqueue_proxy
        while True:
            data = self.request.recv(256)
            if data:
                byte_stream += data
                self.request.sendall(ack)
            else:
                break

        if len(byte_stream) == 0:
            return
        elif byte_stream[0] == PDI_SOP and len(byte_stream) > 0:
            pdi_cmd = PdiCommand.by_value(byte_stream[1], raise_exception=False)
            if pdi_cmd and pdi_cmd.is_sendable:
                # Forward PDI commands to Base3Buffer
                base3_dispatcher: Base3Buffer = cast(ProxyServer, self.server).base3_dispatcher
                if base3_dispatcher is None:
                    base3_dispatcher = cast(ProxyServer, self.server).base3_dispatcher = Base3Buffer.get()
                base3_dispatcher.enqueue_command(byte_stream)
            elif pdi_cmd and pdi_cmd.is_receivable:
                # forward received state command to PdiDispatcher
                pdi_dispatcher: PdiDispatcher = cast(ProxyServer, self.server).pdi_dispatcher
                if pdi_dispatcher is None:
                    pdi_dispatcher = cast(ProxyServer, self.server).pdi_dispatcher = PdiDispatcher.get()
                pdi_dispatcher.offer(byte_stream)
            else:
                log.error(f"Ignoring request to proxy unsendable PDI command: {byte_stream.hex()}")
            return
        # we use TMCC1 syntax to pass special commands to control operating nodes,
        # to reduce overhead, only do the special processing if necessary
        if byte_stream[0] == 0xFE and byte_stream[1] == 0xF0:
            from .command_listener import CommandDispatcher

            # if this is a send state request, deal with it and exit;
            # it has a completely different format than the other
            # sync commands
            if len(byte_stream) > 5 and byte_stream[2] == SENDING_STATE_REQUEST[2]:
                byte_stream = byte_stream[3:]
                CommBuffer.get().update_state(byte_stream)
                return

            # Appended to the admin/sync byte sequence is the port that the server
            # must use to send state updates back to the client. Decode it here
            # client_scope is set if the scope of a command are all the clients
            # on that node ("restart me"). This is only used below iff
            # signal_clients_on is called; this is why we have 2 variables;
            # client_scope & client_ip
            (client_scope, client_port, client_id, client_version) = self.extract_addendum(byte_stream)
            client_ip = client_scope if client_scope else self.client_address[0]
            byte_stream = byte_stream[0:3]
            cmd = CommandReq.from_bytes(byte_stream)

            if byte_stream == DISCONNECT_REQUEST:
                enqueue_proxy.client_disconnect(client_ip, client_port, client_id)
                log.info(f"Client at {client_ip}:{client_port} disconnecting...")
            elif byte_stream == REGISTER_REQUEST:
                if enqueue_proxy.is_client(client_ip, client_port, client_id) is False:
                    ver = f" v{client_version[0]}.{client_version[1]}.{client_version[2]}" if client_version else "?"
                    log.info(f"Client at {client_ip}:{client_port}{ver} connecting...")
                enqueue_proxy.client_connect(client_ip, client_port, client_id)
            elif byte_stream == SYNC_STATE_REQUEST:
                log.info(f"Client at {client_ip}:{client_port} syncing...")
                dispatcher.send_current_state(client_ip, client_port)
            elif byte_stream == KEEP_ALIVE_REQUEST:
                enqueue_proxy.client_alive(client_ip, client_port, client_id)
            elif byte_stream in {
                QUIT_REQUEST,
                REBOOT_REQUEST,
                RESTART_REQUEST,
                SHUTDOWN_REQUEST,
                UPDATE_REQUEST,
                UPGRADE_REQUEST,
            }:
                # admin request, signal all clients
                if client_scope:
                    dispatcher.signal_clients_on(cmd, client_scope)
                else:
                    dispatcher.signal_clients(cmd)
                    dispatcher.publish(CommandScope.SYNC, cmd)
            elif byte_stream == RESYNC_REQUEST:
                dispatcher.publish(CommandScope.SYNC, cmd)
            else:
                log.error(f"Unhandled {cmd} received from {client_ip}:{client_port}")
            # do not send the special PyTrain commands to the Lionel Base 3 or Ser2
            return
        # with the handling of the admin cmds out of the way, queue the bytes
        # received from the client for processing by the Lionel Base 3
        enqueue_proxy.enqueue_request(byte_stream)

    @staticmethod
    def extract_addendum(
        byte_stream: bytes,
    ) -> Tuple[str | None, int | None, uuid.UUID | None, tuple[int, int, int] | None]:
        client_uuid: uuid.UUID | None = None
        client_ip: str | None = None
        client_port: int = DEFAULT_SERVER_PORT
        client_version: tuple[int, int, int] | None = None
        try:
            if len(byte_stream) > 23:  # port and UUID as bytes
                client_port = int.from_bytes(byte_stream[3:5], "big")
                client_uuid = uuid.UUID(bytes=byte_stream[5:21])
                major = byte_stream[21]
                minor = byte_stream[22]
                patch = byte_stream[23]
                client_version = (major, minor, patch)
            elif len(byte_stream) > 18:  # port and UUID as bytes
                client_port = int.from_bytes(byte_stream[3:5], "big")
                client_uuid = uuid.UUID(bytes=byte_stream[5:])
            elif len(byte_stream) > 5:
                addenda = byte_stream[3:].decode("utf-8", errors="ignore")
                parts = addenda.split(":")
                client_ip = parts[0] if parts[0] else None
                client_port = int(parts[1]) if len(parts) > 1 else None
            elif len(byte_stream) > 3:
                client_port = int.from_bytes(byte_stream[3:], "big")
        except Exception as e:
            log.exception(e)
        return client_ip, client_port, client_uuid, client_version



from __future__ import annotations

import sys
from typing import Callable, TypeVar, Set

if sys.version_info >= (3, 11):
    from typing import Self
elif sys.version_info >= (3, 9):
    from typing_extensions import Self

from .constants import (
    DEFAULT_ADDRESS,
    DEFAULT_BAUDRATE,
    DEFAULT_PORT,
    DEFAULT_DURATION_INTERVAL_MSEC,
    MINIMUM_DURATION_INTERVAL_MSEC,
)
from .constants import CommandScope, CommandSyntax
from .tmcc2.tmcc2_constants import TMCC2Enum, TMCC2CommandPrefix, LEGACY_ENGINE_COMMAND_PREFIX
from .tmcc2.tmcc2_constants import TMCC2RouteCommandEnum, TMCC2HaltCommandEnum, TMCC2EngineCommandEnum
from .tmcc2.tmcc2_constants import LEGACY_TRAIN_COMMAND_PREFIX, LEGACY_EXTENDED_BLOCK_COMMAND_PREFIX
from .tmcc2.tmcc2_constants import TMCC2CommandDef
from .command_def import CommandDef, CommandDefEnum
from .tmcc1.tmcc1_constants import TMCC1CommandDef, TMCC1_COMMAND_PREFIX, TMCC1Enum, TMCC1SyncCommandEnum
from .tmcc1.tmcc1_constants import (
    TMCC1HaltCommandEnum,
    TMCC1SwitchCommandEnum,
    TMCC1AuxCommandEnum,
    TMCC1EngineCommandEnum,
)
from .tmcc1.tmcc1_constants import TMCC1CommandIdentifier, TMCC1_TRAIN_COMMAND_PURIFIER
from .tmcc1.tmcc1_constants import TMCC1_TRAIN_COMMAND_MODIFIER
from .tmcc1.tmcc1_constants import TMCC1RouteCommandEnum
from ..utils.validations import Validations

E = TypeVar("E", bound=CommandDefEnum)
R = TypeVar("R", bound="CommandReq")


class CommandReq:
    @classmethod
    def build(
        cls, command: E | bytes, address: int = DEFAULT_ADDRESS, data: int = 0, scope: CommandScope = None
    ) -> Self:
        if isinstance(command, bytes):
            return cls.from_bytes(bytes(command))
        cls._vet_request(command, address, data, scope)
        # we have to do these imports here to avoid cyclic dependencies
        from .sequence.sequence_constants import SequenceCommandEnum
        from ..protocol.multibyte.multibyte_constants import TMCC2MultiByteEnum

        if isinstance(command, SequenceCommandEnum):
            from .sequence.sequence_req import SequenceReq

            return SequenceReq.build(command, address, data, scope)
        elif isinstance(command, TMCC2MultiByteEnum):
            from ..protocol.multibyte.param_command_req import MultiByteReq

            return MultiByteReq.build(command, address, data, scope)
        elif isinstance(command, CommandDefEnum):
            return CommandReq(command, address, data, scope)
        else:
            raise TypeError(f"Command type not recognized {command}")

    @classmethod
    def from_bytes(cls, param: bytes, from_tmcc_rx: bool = False, is_tmcc4: bool = False) -> Self:
        if not param:
            raise ValueError("Command requires at least 3 bytes")
        if len(param) < 3:
            raise ValueError(f"Command requires at least 3 bytes {param.hex(':')}")
        first_byte = int(param[0])
        cmd_req = None
        if first_byte in TMCC4_FIRST_BYTE_TO_INTERPRETER:
            cmd_req = TMCC4_FIRST_BYTE_TO_INTERPRETER[first_byte](param)
        elif is_tmcc4 is False and first_byte in TMCC_FIRST_BYTE_TO_INTERPRETER:
            cmd_req = TMCC_FIRST_BYTE_TO_INTERPRETER[first_byte](param)
        if cmd_req is not None:
            if from_tmcc_rx is True:
                cmd_req._is_tmcc_rx = True
            if is_tmcc4 is True or cmd_req.address > 99:
                cmd_req._is_tmcc4 = True
            return cmd_req
        raise ValueError(f"Command bytes not understood {param.hex(':')}")

    @classmethod
    def send_request(
        cls,
        command: CommandDefEnum,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
        repeat: int = 1,
        delay: float = 0,
        duration: float = 0,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        # build_req & queue
        req = cls.build(command, address, data, scope)
        cls._enqueue_command(req.as_bytes, repeat, delay, duration, baudrate, port, server)

    @classmethod
    def build_action(
        cls,
        command: E | None,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
        repeat: int = 1,
        delay: float = 0,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> Callable:
        # build_req & return action function
        req = cls.build(command, address, data, scope)
        return req.as_action(repeat=repeat, delay=delay, baudrate=baudrate, port=port, server=server)

    @classmethod
    def _determine_first_byte(cls, command: CommandDef, scope: CommandScope) -> bytes:
        """
        Generalized command scopes, such as ENGINE, SWITCH, etc.,
        map to syntax-specific command identifiers defined
        for the TMCC1 and TMCC2 commands
        """
        # otherwise, we need to figure out if we're returning a
        # TMCC1-style or TMCC2-style command prefix
        if isinstance(command, TMCC1CommandDef):
            return TMCC1_COMMAND_PREFIX.to_bytes(1, byteorder="big")
        elif isinstance(command, TMCC2CommandDef):
            validated_scope = cls._validate_requested_scope(command, scope)
            return TMCC2CommandPrefix(validated_scope.name).as_bytes
        raise TypeError(f"Command type not recognized {command}")

    @classmethod
    def _vet_request(
        cls,
        command: CommandDefEnum,
        address: int,
        data: int,
        scope: CommandScope,
    ) -> None:
        from .sequence.sequence_constants import SequenceCommandEnum

        if isinstance(command, TMCC1Enum):
            enum_class = TMCC1Enum
        elif isinstance(command, TMCC2Enum) or isinstance(command, SequenceCommandEnum):
            enum_class = TMCC2Enum
        else:
            raise TypeError(f"Command def not recognized: '{command}'")

        syntax = CommandSyntax.LEGACY if enum_class == TMCC2Enum else CommandSyntax.TMCC
        min_val = 0 if scope in {CommandScope.ENGINE} and syntax == CommandSyntax.LEGACY else 1
        max_val = 9999 if scope in {CommandScope.ENGINE} and syntax == CommandSyntax.LEGACY else 99
        if syntax == CommandSyntax.TMCC and command == TMCC1RouteCommandEnum.FIRE:
            scope = TMCC1CommandIdentifier.ROUTE
            max_val = 31
        if scope is None:
            scope = command.scope
        if command.command_def.is_addressable:
            Validations.validate_int(address, min_value=min_val, max_value=max_val, label=scope.name.title())
        if data is not None and command.command_def.is_data:
            Validations.validate_int(data, label=scope.name.title())

    @classmethod
    def _enqueue_command(
        cls,
        cmd: bytes,
        repeat: int,
        delay: float,
        duration: float,
        baudrate: int,
        port: str | int,
        server: str | None,
        buffer=None,
        request: CommandReq | None = None,
        trigger_effects: bool = True,
        interval: int = None,
    ) -> None:
        repeat = Validations.validate_int(repeat, min_value=1, label="repeat")
        delay = Validations.validate_float(delay, min_value=0, label="delay")
        duration = Validations.validate_float(duration, min_value=0, label="duration", allow_none=True)
        interval = Validations.validate_int(
            interval, min_value=MINIMUM_DURATION_INTERVAL_MSEC, label="interval", allow_none=True
        )

        # send command to comm buffer
        if buffer is None:
            from ..comm.comm_buffer import CommBuffer

            buffer = CommBuffer.build(baudrate=baudrate, port=port, server=server)
        delay = 0 if delay is None else delay
        duration = 0 if duration is None else duration
        for rep_no in range(repeat):
            buffer.enqueue_command(cmd, delay)
            # does this command cause any other state changes?
            if rep_no == 0 and request and trigger_effects is True:
                for effect in cls.results_in(request):
                    if isinstance(effect, CommandDefEnum):
                        effect_cmd = CommandReq.build(effect, request.address, 0, request.scope)
                    elif isinstance(effect, tuple):
                        effect_cmd = CommandReq.build(effect[0], request.address, effect[1], request.scope)
                    else:
                        continue  # we shouldn't ever get here
                    buffer.enqueue_command(effect_cmd.as_bytes, delay)
            if duration > 0:
                # convert duration into milliseconds, then queue a command to fire
                # every 100 msec for the duration
                interval = interval if interval else DEFAULT_DURATION_INTERVAL_MSEC
                for d in range(interval, int(round(duration * 1000)), interval):
                    buffer.enqueue_command(cmd, delay + (d / 1000.0))

    @staticmethod
    def results_in(command: CommandReq) -> Set[E]:
        from ..db.component_state_store import DependencyCache

        dependencies = DependencyCache.build()
        effects = dependencies.results_in(command.command, dereference_aliases=True, include_aliases=False)
        if command.is_data:
            # noinspection PyTypeChecker
            effects.update(
                dependencies.results_in(
                    (command.command, command.data), dereference_aliases=True, include_aliases=False
                )
            )
        # remove the triggering command, and its alias, if it exists
        if effects:
            if command.command in effects:
                effects.remove(command.command)
            if command.is_data and (command.command, command.data) in effects:
                # noinspection PyTypeChecker
                effects.remove((command.command, command.data))
            if command.command.is_alias and command.command.alias in effects:
                alias = command.command.alias
                # noinspection PyTypeChecker
                effects.remove(alias)
                if isinstance(alias, tuple) and alias[0] in effects:
                    effects.remove(alias[0])
        return effects

    @staticmethod
    def _validate_requested_scope(command_def: CommandDef, request: CommandScope) -> CommandScope:
        if request in {CommandScope.ENGINE, CommandScope.TRAIN}:
            if command_def.scope in {CommandScope.ENGINE, CommandScope.TRAIN}:
                return request
        # otherwise, return the scope associated with the native command def
        return command_def.scope

    def __init__(
        self,
        command_def_enum: E,
        address: int = DEFAULT_ADDRESS,
        data: int = 0,
        scope: CommandScope = None,
    ) -> None:
        self._command_def_enum: E = command_def_enum
        self._command_name = self._command_def_enum.name
        # noinspection PyTypeChecker
        self._command_def: TMCC2CommandDef = command_def_enum.value  # read-only; do not modify
        if self._command_def.is_addressable:
            self._address = address
        else:
            self._address = 0
        if self._command_def.is_data:
            self._data = data
        else:
            self._data = 0
        self._native_scope = self._command_def.scope
        self._scope = self._validate_requested_scope(self._command_def, scope)
        from ..comm.comm_buffer import CommBuffer

        self._buffer: CommBuffer | None = None
        self._message_processor = None
        self._is_tmcc_rx = False
        self._is_tmcc4 = False

        # save the command bits from the def, as we will be modifying them
        self._command_bits: int = self._command_def.bits

        # apply the given address and data
        self._apply_address()
        self._apply_data()

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __call__(self, message: CommandReq) -> None:
        """
        Allows CommandReqs to receive messages from channels
        this request is subscribed to
        """
        if self._message_processor:
            self._message_processor(message)

    def __repr__(self) -> str:
        if self.is_data:
            data = f" {self.data}"
        else:
            data = ""
        rx = " (RX)" if self.is_tmcc_rx else ""
        return f"[{self.scope.name} {self.address} {self.command_name}{data}{rx} (0x{self.as_bytes.hex()})]"

    @property
    def address(self) -> int:
        if self.command_def.is_addressable:
            return self._address
        else:
            return DEFAULT_ADDRESS

    @address.setter
    def address(self, new_address: int) -> None:
        if self.command_def.is_addressable and new_address != self._address:
            self._address = new_address
            self._apply_address()

    @property
    def data(self) -> int:
        return self._data

    @data.setter
    def data(self, new_data: int) -> None:
        if self.is_data and new_data != self._data:
            self._data = new_data
            self._apply_data()

    @property
    def scope(self) -> CommandScope:
        return self._scope

    @scope.setter
    def scope(self, new_scope: CommandScope) -> None:
        if new_scope != self._scope:
            # can only change scope for Engine and Train commands, and then, just from the one to the other
            if self.scope in {CommandScope.ENGINE, CommandScope.TRAIN}:
                if new_scope in {CommandScope.ENGINE, CommandScope.TRAIN}:
                    self._scope = new_scope
                    self._apply_scope()
                    return
            raise AttributeError(f"Scope {new_scope} not supported for {self}")

    @property
    def is_halt(self) -> bool:
        return self.command == TMCC1HaltCommandEnum.HALT

    @property
    def is_system_halt(self) -> bool:
        return self.command in [TMCC2HaltCommandEnum.HALT, TMCC2EngineCommandEnum.SYSTEM_HALT]

    @property
    def command(self) -> E:
        return self._command_def_enum

    @property
    def command_name(self) -> str:
        return self._command_name

    @property
    def command_def(self) -> TMCC2CommandDef:
        return self._command_def

    @property
    def native_scope(self) -> CommandScope:
        return self._native_scope

    @property
    def bits(self) -> int:
        return self._command_bits

    @property
    def is_data(self) -> bool:
        return self.command_def.num_data_bits != 0

    @property
    def num_data_bits(self) -> int:
        return self.command_def.num_data_bits

    @property
    def data_max(self) -> int:
        return self.command_def.data_max

    @property
    def data_min(self) -> int:
        return self.command_def.data_min

    @property
    def syntax(self) -> CommandSyntax:
        return self._command_def.syntax

    @property
    def is_tmcc1(self) -> bool:
        return self._command_def.is_tmcc1

    @property
    def is_tmcc2(self) -> bool:
        return self._command_def.is_tmcc2

    @property
    def identifier(self) -> int | None:
        return self.command_def.identifier

    @property
    def is_tmcc_rx(self) -> bool:
        return self._is_tmcc_rx

    @property
    def is_filtered(self) -> bool:
        return self.command_def.is_filtered is True and self.is_tmcc_rx is False

    @property
    def is_force_state_update(self) -> bool:
        """
        For the most part, the Base 3 doesn't broadcast received commands to other receivers.
        This prevents the PyTrain server and attached clients from maintaining accurate state
        for 4-digit engines. To compensate, we echo all commands to 4-digit engines after we
        send them to the Base 3.

        This method is used to determine if a command should be force-broadcast to all clients.
        """
        if isinstance(self.command_def, TMCC2CommandDef):
            return self.command_def.is_d4_broadcast is False and self.address > 99
        return False

    def send(
        self,
        repeat: int = 1,
        delay: float = 0,
        duration: float = 0,
        interval: int = None,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        interval = self.command_def.interval if self.command_def.interval else interval
        self._enqueue_command(
            self.as_bytes,
            repeat,
            delay,
            duration,
            baudrate,
            port,
            server,
            request=self,
            interval=interval,
        )

    @property
    def as_bytes(self) -> bytes:
        if self.scope is None:
            first_byte = self.command_def.first_byte
        else:
            first_byte = self._determine_first_byte(self.command_def, self.scope)
        byte_str = first_byte + self._command_bits.to_bytes(2, byteorder="big")
        if self.address > 99:
            byte_str += str(self.address).zfill(4).encode()
        return byte_str

    def as_action(
        self,
        repeat: int = 1,
        delay: float = 0,
        duration: float = 0,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
        address: int = None,
        data: int = None,
    ) -> Callable:
        from ..comm.comm_buffer import CommBuffer

        buffer = CommBuffer.build(baudrate=baudrate, port=port, server=server)

        def send_func(new_address: int = None, new_data: int = None, trigger_effects: bool = True) -> None:
            if new_address and new_address != self.address:
                self.address = new_address
            elif address and address != self.address:
                self.address = address
            if self.num_data_bits:
                if new_data is not None and new_data != self.data:
                    self.data = new_data
                elif data is not None and data != self.data:
                    self.data = data

            self._enqueue_command(
                self.as_bytes,
                repeat=repeat,
                delay=delay,
                duration=duration,
                baudrate=baudrate,
                port=port,
                server=server,
                buffer=buffer,
                request=self,
                trigger_effects=trigger_effects,
            )

        return send_func

    def _apply_address(self, new_address: int = None) -> int:
        if not self.command_def.is_addressable:  # HALT command
            return self._command_bits
        # reset existing address bits, if any
        self._command_bits &= self._command_def.address_mask
        # figure out which address we're using
        the_address = new_address if new_address and new_address > 0 else self._address
        if self.syntax == CommandSyntax.TMCC:
            self._command_bits |= the_address << 7
            if self.scope == CommandScope.TRAIN and self.identifier == TMCC1CommandIdentifier.ENGINE:
                self._command_bits &= TMCC1_TRAIN_COMMAND_PURIFIER
                self._command_bits |= TMCC1_TRAIN_COMMAND_MODIFIER
        elif self.syntax == CommandSyntax.LEGACY:
            if 1 <= the_address <= 99:
                self._command_bits |= the_address << 9
        else:
            raise ValueError(f"Command syntax not recognized {self.syntax}")
        return self._command_bits

    def _apply_data(self, new_data: int = None) -> int:
        """
        For commands that take parameters, such as engine speed and brake level,
        apply the data bits to the command action bytes to form the complete byte
        set to send to the Lionel LCS SER2.
        """
        data = new_data if new_data is not None else self.data
        if self.num_data_bits and data is None:
            raise ValueError("Data is required")
        if self.num_data_bits == 0:
            return self.bits
        elif self.command_def.data_map:
            if data in self.command_def.data_map:
                data = self.command_def.data_map[data]
            else:
                raise ValueError(f"Invalid data: {data} (not in map)")
        elif data < self.command_def.data_min or data > self.command_def.data_max:
            raise ValueError(f"Invalid: {data} not in range {self.command_def.data_min}-{self.command_def.data_max}")
        # sanitize data so we don't set bits we shouldn't
        data_bits = 2**self.num_data_bits - 1
        filtered_data = data & data_bits
        if data != filtered_data:
            raise ValueError(f"Invalid data: {data} (not in range)")
        # clear out old data
        self._command_bits &= 0xFFFF & ~data_bits
        # set new data
        self._command_bits |= data
        return self._command_bits

    def _apply_scope(self, new_scope: CommandScope = None) -> int:
        scope = new_scope if new_scope is not None else self.scope
        if self.syntax == CommandSyntax.TMCC and self.identifier == TMCC1CommandIdentifier.ENGINE:
            self._command_bits &= TMCC1_TRAIN_COMMAND_PURIFIER
            if scope == CommandScope.TRAIN:
                self._command_bits |= TMCC1_TRAIN_COMMAND_MODIFIER
        elif self.syntax == CommandSyntax.LEGACY:
            pass
        else:
            raise ValueError(f"Command syntax not recognized {self.syntax}")
        return self._command_bits

    # noinspection PyUnresolvedReferences
    @classmethod
    def build_tmcc1_command_req(cls, param: bytes) -> Self:
        value = int.from_bytes(param[1:3], byteorder="big")
        for tmcc_enum in [
            TMCC1HaltCommandEnum,
            TMCC1SwitchCommandEnum,
            TMCC1AuxCommandEnum,
            TMCC1RouteCommandEnum,
            TMCC1EngineCommandEnum,
            TMCC1SyncCommandEnum,
        ]:
            scope = None
            cmd_enum = tmcc_enum.by_value(value)
            if (
                cmd_enum is None
                and tmcc_enum == TMCC1EngineCommandEnum
                and (value & TMCC1_TRAIN_COMMAND_MODIFIER) == TMCC1_TRAIN_COMMAND_MODIFIER
            ):
                # check if this is a TRAIN command and if so, clear out the
                # train bits and look again; only do this for engine commands
                cmd_enum = tmcc_enum.by_value(value & TMCC1_TRAIN_COMMAND_PURIFIER)
                if cmd_enum:
                    scope = CommandScope.TRAIN
            if cmd_enum:
                # build_req the request and return
                data = cmd_enum.value.data_from_bytes(param[1:3])
                address = cmd_enum.value.address_from_bytes(param[1:3])
                return CommandReq.build(cmd_enum, address, data, scope)
        raise ValueError(f"Invalid tmcc1 command: {param.hex(':')}")

    @classmethod
    def build_tmcc2_command_req(cls, param: bytes, is_tmcc4: bool = False) -> R:
        if len(param) == 3 or (len(param) == 7 and param[1] in {0x00, 0x01}):
            value = int.from_bytes(param[1:3], byteorder="big")
            for tmcc_enum in [TMCC2HaltCommandEnum, TMCC2EngineCommandEnum, TMCC2RouteCommandEnum]:
                cmd_enum = tmcc_enum.by_value(value)
                if cmd_enum is not None:
                    scope = cmd_enum.scope
                    if int(param[0]) == LEGACY_TRAIN_COMMAND_PREFIX:
                        scope = CommandScope.TRAIN
                    # build_req the request and return
                    data = cmd_enum.value.data_from_bytes(param[1:3])
                    if len(param) == 7 and param[1] in {0x00, 0x01}:
                        # TODO: this code looks fragile; should rethink
                        addr_str = ""
                        for i in range(3, 7):
                            addr_str += chr(param[i])
                        address = int(addr_str)
                    else:
                        address = cmd_enum.value.address_from_bytes(param[1:3])
                    return CommandReq.build(cmd_enum, address, data, scope)
            raise ValueError(f"Invalid tmcc2 command: {param.hex(':')}")
        else:
            from ..protocol.multibyte.multibyte_command_req import MultiByteReq

            return MultiByteReq.from_bytes(param, is_tmcc4=is_tmcc4)

    @classmethod
    def build_tmcc4_command_req(cls, param: bytes) -> R:
        return cls.build_tmcc2_command_req(param, is_tmcc4=True)


TMCC4_FIRST_BYTE_TO_INTERPRETER = {
    LEGACY_ENGINE_COMMAND_PREFIX: CommandReq.build_tmcc4_command_req,
}

TMCC_FIRST_BYTE_TO_INTERPRETER = {
    TMCC1_COMMAND_PREFIX: CommandReq.build_tmcc1_command_req,
    LEGACY_ENGINE_COMMAND_PREFIX: CommandReq.build_tmcc2_command_req,
    LEGACY_TRAIN_COMMAND_PREFIX: CommandReq.build_tmcc2_command_req,
    LEGACY_EXTENDED_BLOCK_COMMAND_PREFIX: CommandReq.build_tmcc2_command_req,
}



#!/usr/bin/env python3

#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import logging
from argparse import ArgumentParser
from typing import List

from . import CliBaseTMCC
from ..protocol.multibyte.effects_cmd import EffectsCmd
from ..protocol.multibyte.multibyte_constants import TMCC2EffectsControl
from ..utils.argument_parser import PyTrainArgumentParser

log = logging.getLogger(__name__)


class EffectsCli(CliBaseTMCC):
    @classmethod
    def command_parser(cls):
        effects_parser = PyTrainArgumentParser(add_help=False)
        effects_parser.add_argument("engine", metavar="Engine/Train", type=int, help="Engine/Train to control")
        sp = effects_parser.add_subparsers(dest="sub_command", help="Engine/train sub-commands")
        smoke = sp.add_parser("smoke", aliases=["sm"], help="Smoke system options")
        smoke_group = smoke.add_mutually_exclusive_group()
        smoke_group.add_argument(
            "-low",
            action="store_const",
            const=TMCC2EffectsControl.SMOKE_LOW,
            dest="option",
            default=TMCC2EffectsControl.SMOKE_LOW,
            help="Smoke system low",
        )
        smoke_group.add_argument(
            "-medium",
            action="store_const",
            const=TMCC2EffectsControl.SMOKE_MEDIUM,
            dest="option",
            help="Smoke system medium",
        )
        smoke_group.add_argument(
            "-high",
            "-maximum",
            action="store_const",
            const=TMCC2EffectsControl.SMOKE_HIGH,
            dest="option",
            help="Smoke system high",
        )
        smoke_group.add_argument(
            "-off", action="store_const", const=TMCC2EffectsControl.SMOKE_OFF, dest="option", help="Smoke system off"
        )

        panto = sp.add_parser("pantograph", aliases=["pa"], help="Pantograph options")
        panto_group = panto.add_mutually_exclusive_group()
        panto_group.add_argument(
            "-fu",
            "--front_up",
            action="store_const",
            const=TMCC2EffectsControl.PANTO_FRONT_UP,
            dest="option",
            default=TMCC2EffectsControl.PANTO_FRONT_UP,
            help="Pantograph control - front up",
        )
        panto_group.add_argument(
            "-fd",
            "--front_down",
            action="store_const",
            const=TMCC2EffectsControl.PANTO_FRONT_DOWN,
            dest="option",
            help="Pantograph control - front down",
        )
        panto_group.add_argument(
            "-bu",
            "--both_up",
            action="store_const",
            const=TMCC2EffectsControl.PANTO_BOTH_UP,
            dest="option",
            help="Pantograph control -both up",
        )
        panto_group.add_argument(
            "-bd",
            "--both_down",
            action="store_const",
            const=TMCC2EffectsControl.PANTO_BOTH_DOWN,
            dest="option",
            help="Pantograph control - both down",
        )
        panto_group.add_argument(
            "-ru",
            "--rear_up",
            action="store_const",
            const=TMCC2EffectsControl.PANTO_REAR_UP,
            dest="option",
            help="Pantograph control - rear up",
        )
        panto_group.add_argument(
            "-rd",
            "--rear_down",
            action="store_const",
            const=TMCC2EffectsControl.PANTO_REAR_DOWN,
            dest="option",
            help="Pantograph control - rear down",
        )

        subway = sp.add_parser("subway", aliases=["su"], help="Subway door options")
        subway_group = subway.add_mutually_exclusive_group()
        subway_group.add_argument(
            "-lo",
            "--left_open",
            action="store_const",
            const=TMCC2EffectsControl.SUBWAY_LEFT_DOOR_OPEN,
            dest="option",
            default=TMCC2EffectsControl.SUBWAY_LEFT_DOOR_OPEN,
            help="Subway left door open",
        )
        subway_group.add_argument(
            "-lc",
            "--left_close",
            action="store_const",
            const=TMCC2EffectsControl.SUBWAY_LEFT_DOOR_CLOSE,
            dest="option",
            help="Subway left door close",
        )
        subway_group.add_argument(
            "-bo",
            "--both_open",
            action="store_const",
            const=TMCC2EffectsControl.SUBWAY_BOTH_DOOR_OPEN,
            dest="option",
            help="Subway both door open",
        )
        subway_group.add_argument(
            "-bc",
            "--both_close",
            action="store_const",
            const=TMCC2EffectsControl.SUBWAY_BOTH_DOOR_CLOSE,
            dest="option",
            help="Subway both door close",
        )
        subway_group.add_argument(
            "-ro",
            "--right_open",
            action="store_const",
            const=TMCC2EffectsControl.SUBWAY_RIGHT_DOOR_OPEN,
            dest="option",
            help="Subway right door open",
        )
        subway_group.add_argument(
            "-rc",
            "--right_close",
            action="store_const",
            const=TMCC2EffectsControl.SUBWAY_RIGHT_DOOR_CLOSE,
            dest="option",
            help="Subway right door close",
        )
        # TODO: Coal and stock car options

        return PyTrainArgumentParser(
            "Engine/train/car effects control",
            parents=[effects_parser, cls.multi_parser(), cls.train_parser(), cls.cli_parser()],
        )

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        engine: int = self._args.engine
        option = self._args.option
        try:
            scope = self._determine_scope()
            cmd = EffectsCmd(
                engine,
                TMCC2EffectsControl(option),
                0,
                scope,
                baudrate=self._baudrate,
                port=self._port,
                server=self._server,
            )
            if self.do_fire:
                cmd.fire(baudrate=self._baudrate, port=self._port, server=self._server)
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)



from threading import RLock, Thread
from time import sleep
from typing import Callable, Dict

from gpiozero import RotaryEncoder

from ..comm.comm_buffer import CommBuffer
from ..gpio.gpio_handler import GpioHandler, log
from ..protocol.command_req import CommandReq
from ..protocol.constants import PROGRAM_NAME


class PyRotaryEncoder(RotaryEncoder):
    def __init__(
        self,
        pin_1: int | str,
        pin_2: int | str,
        command: CommandReq = None,
        wrap: bool = True,
        max_steps: int = 100,
        initial_step: int = 0,
        ramp: Dict[int, int] = None,
        steps_to_data: Callable[[int], int] = None,
        data_to_steps: Callable[[int], int] = None,
        pause_for: float = 0.25,
        reset_after_motion: bool = False,
    ):
        super().__init__(pin_1, pin_2, wrap=wrap, max_steps=max_steps)
        if initial_step is not None:
            self.steps = initial_step
        self._command = command
        self._action = command.as_action() if command else None
        self._ramp = ramp
        self._steps_to_data = steps_to_data
        self._data_to_steps = data_to_steps
        self._tmcc_command_buffer = CommBuffer.build()
        self._last_known_data = None
        self._last_rotation_at = GpioHandler.current_milli_time()
        self._last_steps = self.steps
        self._lock = RLock()
        self._pins = (pin_1, pin_2)
        self._handler = PyRotaryEncoderHandler(self, pause_for=pause_for, reset_after_motion=reset_after_motion)
        if command:
            self._handler.start()

    @property
    def pins(self) -> tuple[int, int]:
        return self._pins

    @property
    def lock(self) -> RLock:
        return self._lock

    def update_action(
        self,
        command: CommandReq,
        state,
        steps_to_data: Callable[[int], int],
        data_to_steps: Callable[[int], int],
    ) -> None:
        self._steps_to_data = steps_to_data
        self._data_to_steps = data_to_steps
        cur_speed = state.speed if state and state.speed is not None else 0
        self.steps = self._last_steps = data_to_steps(cur_speed)
        self._command = command
        self._action = command.as_action() if command else None
        self._handler.start()

    def update_data(self, new_data) -> None:
        if new_data != self._last_known_data and self._data_to_steps and not self.is_active:
            with self.lock:
                log.debug(f"{self._last_known_data} -> {self._data_to_steps(new_data)}")
                self.steps = self._data_to_steps(new_data)
                self._last_known_data = new_data
                if self._handler:
                    self._handler.reset_last_known()

    def fire_action(self, cur_step: int = None) -> None:
        if self._action:
            if cur_step is None:
                cur_step = self.steps
            if self._steps_to_data:
                self._last_known_data = data = self._steps_to_data(cur_step)
            else:
                data = 0  # TODO: handle this case
            self._action(new_data=data)

    @property
    def last_rotation_at(self) -> int:
        """
        Return the time of the last rotation in millisecond resolution
        """
        return self._last_rotation_at

    @property
    def last_rotated(self) -> float:
        """
        Seconds since the last rotation.
        """
        return (GpioHandler.current_milli_time() - self.last_rotation_at) / 1000.0

    def reset(self) -> None:
        if self._handler:
            self._handler.reset()


class PyRotaryEncoderHandler(Thread):
    def __init__(self, re: PyRotaryEncoder, pause_for: float = 0.25, reset_after_motion: bool = False) -> None:
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Rotary Encoder Handler pins: {re.pins}")
        self._re = re
        self._command = None
        self._is_running = True
        self._is_started = False
        self._last_step = None
        self._pause_for = pause_for
        self._reset_after_motion = reset_after_motion
        if reset_after_motion is True:
            self._re.steps = 0

    def start(self) -> None:
        with self._re.lock:
            if self._is_started is False:
                GpioHandler.cache_handler(self)
                super().start()

    def reset_last_known(self) -> None:
        with self._re.lock:
            self._last_step = self._re.steps

    def run(self) -> None:
        with self._re.lock:
            self._is_started = True
        self._last_step = 0 if self._reset_after_motion is True else float("-inf")
        # In general, Rotary Encoder is used to control speed. When the encoder steps is
        # -max_steps, this usually corresponds with speed 0; make sure speed zero cmds
        # are sent
        num_neg_max_steps = 3
        while self._is_running:
            with self._re.lock:
                cur_step = self._re.steps
            if cur_step != self._last_step or (num_neg_max_steps > 0 and cur_step == -self._re.max_steps):
                if cur_step == -self._re.max_steps:
                    num_neg_max_steps -= 1
                else:
                    num_neg_max_steps = 3
                self._re.fire_action(cur_step)
                if self._reset_after_motion:
                    self._re.steps = cur_step = 0
                self._last_step = cur_step
            sleep(self._pause_for)

    def reset(self) -> None:
        self._is_running = False



from __future__ import annotations

from .constants import PdiCommand, PDI_SOP, PDI_EOP
from .pdi_req import PdiReq
from ..atc.block import Block
from ..db.block_state import BlockState
from ..protocol.constants import CommandScope, Direction


class BlockReq(PdiReq):
    def __init__(
        self,
        data: bytes | Block | BlockState,
        pdi_command: PdiCommand = PdiCommand.BLOCK_RX,
    ) -> None:
        super().__init__(data, pdi_command)
        self._scope = CommandScope.BLOCK
        if isinstance(data, bytes):
            data_len = len(self._data)
            self._block_id = self._tmcc_id = self._data[1] if data_len > 0 else None
            self._prev_block_id = self._data[2] if data_len > 2 else None
            self._next_block_id = self._data[3] if data_len > 3 else None
            self._flags = self._data[4] if data_len > 4 else None
            self._sensor_track_id = self._data[5] if data_len > 5 else None
            self._switch_id = self._data[6] if data_len > 6 else None
            self._motive_id = int.from_bytes(self._data[7:9], byteorder="little") if data_len > 7 else None
            self._motive_scope = CommandScope.by_value(self._data[9]) if data_len > 9 else None
            self._motive_direction = Direction.by_value(self._data[10]) if data_len > 10 else None
            self._name = self.decode_text(self._data[11:44]) if data_len > 11 else None
            self._direction = Direction.L2R if (self.flags & 0x10 > 0) else Direction.R2L
        elif isinstance(data, Block):
            self._block_id = self._tmcc_id = data.block_id
            self._prev_block_id = data.prev_block.block_id if data.prev_block else None
            self._next_block_id = data.next_block.block_id if data.next_block else None
            self._sensor_track_id = data.sensor_track.address if data.sensor_track else None
            self._switch_id = data.switch.address if data.switch else None
            self._name = data.name
            self._direction = data.direction
            if data.occupied_by:
                self._motive_id = data.occupied_by.address
                self._motive_scope = data.occupied_by.scope
                self._motive_direction = data.occupied_direction
            else:
                self._motive_id = None
                self._motive_scope = None
                self._motive_direction = None
            flags = 0
            for i, flag in enumerate(
                [data.is_occupied, data.is_entered, data.is_slowed, data.is_stopped, data.is_left_to_right]
            ):
                flags |= (1 << i) if flag is True else 0
            self._flags = flags
        elif isinstance(data, BlockState):
            self._block_id = self._tmcc_id = data.block_id
            self._prev_block_id = data.prev_block.block_id if data.prev_block else None
            self._next_block_id = data.next_block.block_id if data.next_block else None
            self._sensor_track_id = data.sensor_track.address if data.sensor_track else None
            self._switch_id = data.switch.address if data.switch else None
            self._name = data.name
            self._direction = data.direction
            if data.occupied_by:
                self._motive_id = data.occupied_by.address
                self._motive_scope = data.occupied_by.scope
                self._motive_direction = data.occupied_direction
            else:
                self._motive_id = None
                self._motive_scope = None
                self._motive_direction = None
            self._flags = data.flags
        else:
            raise AttributeError(f"Unsupported argument: {data}")

    @property
    def payload(self) -> str | None:
        oc = f" Occupied: {self.is_occupied}"
        nm = f"  {self.name}" if self.name else ""
        return f"Block Id: {self.block_id}{nm}{oc} {self.packet}"

    @property
    def as_bytes(self) -> bytes:
        if self._original:
            return self._original
        byte_str = self.pdi_command.as_bytes
        byte_str += self.block_id.to_bytes(1, byteorder="big")
        byte_str += (self._prev_block_id if self._prev_block_id else 0).to_bytes(1, byteorder="big")
        byte_str += (self._next_block_id if self._next_block_id else 0).to_bytes(1, byteorder="big")
        byte_str += (self._flags if self._flags else 0).to_bytes(1, byteorder="big")
        byte_str += (self._sensor_track_id if self._sensor_track_id else 0).to_bytes(1, byteorder="big")
        byte_str += (self._switch_id if self._switch_id else 0).to_bytes(1, byteorder="big")
        byte_str += (self._motive_id if self._motive_id else 0).to_bytes(2, byteorder="little")
        byte_str += (self._motive_scope.value if self._motive_scope else 0).to_bytes(1, byteorder="big")
        byte_str += (self._motive_direction.value if self._motive_direction else 0).to_bytes(1, byteorder="big")
        byte_str += self.encode_text(self._name, 33)
        byte_str, checksum = self._calculate_checksum(byte_str)
        byte_str = PDI_SOP.to_bytes(1, byteorder="big") + byte_str
        byte_str += checksum
        byte_str += PDI_EOP.to_bytes(1, byteorder="big")
        return byte_str

    @property
    def block_id(self) -> int:
        return self._block_id

    @property
    def name(self) -> str:
        return self._name

    @property
    def direction(self) -> Direction:
        return self._direction

    @property
    def prev_block_id(self) -> int:
        return self._prev_block_id

    @property
    def next_block_id(self) -> int:
        return self._next_block_id

    @property
    def sensor_track_id(self) -> int:
        return self._sensor_track_id

    @property
    def switch_id(self) -> int:
        return self._switch_id

    @property
    def motive_id(self) -> int:
        return self._motive_id

    @property
    def motive_scope(self) -> CommandScope:
        return self._motive_scope

    @property
    def motive_direction(self) -> Direction:
        return self._motive_direction

    @property
    def flags(self) -> int:
        return self._flags

    @property
    def is_occupied(self) -> bool:
        return self._flags & (1 << 0) != 0



from src.pytrain import get_version
from .test_base import TestBase


class TestPyTrain(TestBase):
    def test_get_version(self):
        assert get_version() is not None
        assert get_version().startswith("v")



from .multibyte_constants import TMCC2RailSoundsDialogControl
from ..tmcc2.tmcc2_command import TMCC2FixedParameterCommand
from ..constants import CommandScope, DEFAULT_BAUDRATE, DEFAULT_PORT


class DialogCmd(TMCC2FixedParameterCommand):
    def __init__(
        self,
        engine: int,
        command: TMCC2RailSoundsDialogControl,
        data: int = 0,
        scope: CommandScope = CommandScope.ENGINE,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        super().__init__(command, engine, data, scope, baudrate, port, server)
